from __future__ import annotations

import hashlib
import json
import logging
import os
import shutil
from io import BufferedReader, BytesIO
from typing import Union
from uuid import UUID

from django.conf import settings
from django.core.files.uploadedfile import (
    InMemoryUploadedFile,
    TemporaryUploadedFile,
)
from django.db.models.fields.files import FieldFile, ImageFieldFile
from django.utils import timezone


FORMAT = "%(levelname)s:%(asctime)s:%(name)s:%(funcName)s:%(lineno)d >>> %(message)s"
DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


def _attr_to_str(*args, **kwargs) -> str:
    return ":".join([str(arg) for arg in args] + [json.dumps(kwargs, default=str)])


def get_logger(name: str = __name__, level: int | str = logging.INFO) -> logging.Logger:
    logging.basicConfig(format=FORMAT, datefmt=DATE_FORMAT, level=level)
    logger = logging.getLogger(name)
    logger.setLevel(level)
    return logger


def make_md5_hash(*args, **kwargs) -> str:
    return hashlib.md5(_attr_to_str(*args, **kwargs).encode("utf-8")).hexdigest()


def make_uuid(*args, **kwargs) -> UUID:
    return UUID(hex=make_md5_hash(*args, **kwargs))


def remove_dir(dir_path: str) -> None:
    try:
        shutil.rmtree(dir_path)
    except OSError:
        pass


def create_dir(dir_path: str) -> None:
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


def get_dir(upload_to: str = None) -> str:
    media_root = str(settings.MEDIA_ROOT) if settings.MEDIA_ROOT else ""
    root_dir = os.path.join(
        media_root,
        (
            timezone.now().strftime(upload_to)
            if isinstance(upload_to, str)
            else timezone.now().strftime("%Y/%m/%d")
        ),
    )
    return root_dir


def safe_remove_file(fp: str) -> None:
    try:
        os.remove(fp)
    except OSError:
        pass


def get_filename(fp: str, remove_extension: bool = False) -> str:
    _, filename = os.path.split(fp)
    if remove_extension:
        filename = os.path.splitext(filename)[0]
    return filename


def get_file_extension(fp: str) -> str:
    _, extension = os.path.splitext(fp)
    return extension


def get_save_file_path(fp: str, upload_to: str = None) -> str:
    filename = get_filename(fp)
    root_dir = get_dir(upload_to)
    create_dir(root_dir)
    fp = os.path.join(root_dir, filename)
    return fp


def get_paths(fp: str, upload_to: str = "") -> tuple[str, str]:
    save_fp = get_save_file_path(fp, upload_to)
    media_root = str(settings.MEDIA_ROOT) if settings.MEDIA_ROOT else ""
    if media_root:
        fp = save_fp.split(media_root)[-1]
    else:
        fp = save_fp

    if fp[0] == "/":
        fp = fp[1:]

    return save_fp, fp


def get_file_path(filename: str, upload_dir: str = "") -> str:
    _save_fp, fp = get_paths(filename, upload_dir)
    return fp


def join_file_path(*args: str) -> str:
    return os.path.join(*args)


def handle_upload_file(file, upload_dir: str = None):
    save_fp, fp = get_paths(file.name, upload_dir)
    with open(save_fp, "wb+") as f:
        for chunk in file.chunks():
            f.write(chunk)
        return fp


def get_md5_checksum(
    fp: Union[
        str,
        bytes,
        BytesIO,
        BufferedReader,
        InMemoryUploadedFile,
        TemporaryUploadedFile,
        FieldFile,
        ImageFieldFile,
    ],
    chunk_size: int = 65536,
    closed: bool = True,
) -> str:
    md5hash = hashlib.md5()
    if isinstance(fp, (InMemoryUploadedFile, TemporaryUploadedFile)):
        for chunk in fp.chunks(chunk_size):
            md5hash.update(chunk)
    else:
        if isinstance(fp, str):
            fp = open(fp, "rb")
        elif isinstance(fp, (FieldFile, ImageFieldFile)):
            fp = fp.file.file
        elif isinstance(fp, bytes):
            fp = BytesIO(fp)

        while chunk := fp.read(chunk_size):
            md5hash.update(chunk)

        if closed and not fp.closed:
            fp.close()

    return md5hash.hexdigest()



from __future__ import annotations

from io import BufferedReader, BytesIO
from typing import TYPE_CHECKING, Any, BinaryIO, Optional, Union
from uuid import UUID

from django.db.models.fields.files import FieldFile, ImageFieldFile

from PIL import Image, ImageFile, UnidentifiedImageError
from PIL.JpegImagePlugin import JpegImageFile
from PIL.PngImagePlugin import PngImageFile
from PIL.WebPImagePlugin import WebPImageFile

from .app_settings import app_settings
from .constants import TypeChoices
from .utils import get_logger, get_md5_checksum, get_paths, safe_remove_file


if TYPE_CHECKING:
    from .models import FileManager
    from .typed import File

LOGGER = get_logger(__name__)

_File = Union[str, bytes, BytesIO, BufferedReader, FieldFile, ImageFieldFile]
_OpenFile = Union[BinaryIO, BytesIO, BufferedReader, Any]
_Image = Union[ImageFile.ImageFile, JpegImageFile, PngImageFile, WebPImageFile]
_ImageFile = Union[_File, _Image]


class BaseOptimizer:
    """Base Optimizer"""

    def __init__(self, instance: FileManager, file: File, *args, **kwargs):
        self._instance = instance
        self._file = file

    @property
    def instance(self) -> FileManager:
        return self._instance

    @property
    def file(self) -> File:
        return self._file

    @classmethod
    def open(cls, fp: _File) -> None | BufferedReader | BytesIO:
        return cls._open(fp)

    @classmethod
    def _open(cls, fp: _File) -> _OpenFile:
        if isinstance(fp, _File):
            if isinstance(fp, str):
                return open(fp, "rb")
            if isinstance(fp, bytes):
                fp = BytesIO(fp)
            if isinstance(fp, (BytesIO, BufferedReader)):
                return fp
            return fp.file.file

    @classmethod
    def close(cls, fp: _File) -> None:
        if isinstance(fp, (FieldFile, ImageFieldFile)):
            fp.close()
        else:
            if (
                fp
                and getattr(fp, "close", None)
                and bool(getattr(fp, "closed", None) is False)
            ):
                fp.close()

    @classmethod
    def checksum(cls, fp: _File):
        return get_md5_checksum(fp)

    @classmethod
    def get_identifier(cls, fp: _File) -> UUID:
        return UUID(hex=cls.checksum(fp))

    def run(self):
        pass


class ImageOptimizer(BaseOptimizer):
    """Image Optimizer"""

    _supported_file_types = (".jpg", ".jpeg", ".png", ".webp")

    def __init__(
        self,
        instance,
        file,
        *args,
        **kwargs,
    ):
        super().__init__(instance, file, *args, **kwargs)

    def run(self):
        image, path = self.optimize(self.file.save_path, upload_to=self.file._upload_to)
        self.file.path = path
        self.close(image)

    @classmethod
    def open(cls, fp: _File) -> _Image:
        try:
            if isinstance(fp, _Image):
                return fp

            image = Image.open(cls._open(fp))
            return image
        except (UnidentifiedImageError, FileNotFoundError):
            pass

    @classmethod
    def close(cls, fp: _ImageFile) -> None:
        if isinstance(fp, (Image.Image, FieldFile, ImageFieldFile)):
            fp.close()
        else:
            super().close(fp)

    @classmethod
    def optimize(
        cls,
        fp: _ImageFile,
        *,
        filename: str = None,
        upload_to: str = None,
        box: tuple[int, int, int, int] = None,
        max_width: int = app_settings.image_optimizer.max_width,
        max_height: int = app_settings.image_optimizer.max_height,
        to_webp: bool = app_settings.image_optimizer.to_webp,
        remove_origin: bool = app_settings.image_optimizer.remove_origin,
        quality: int = app_settings.image_optimizer.quality,
        compress_level: int = app_settings.image_optimizer.compress_level,
    ) -> tuple[Optional[_Image], Optional[str]]:
        """Optimize the Image File

        Args:
          fp: File path or file object.
          filename: Rename the original file.
          upload_to: Upload dir.
          box: The crop rectangle, as a (left, upper, right, lower)-tuple to crop the image.
          max_width: Max width of the image to resize.
          max_height: Max height of the image to resize.
          to_webp: Force convert image to webp type.
          remove_origin: Force to delete original image after optimization.
          quality: Quality of image.
          compress_level: Compress level to optimize (from 1 - 9).

        Returns:
          The Tuple: PIL Image, Image file path location. If the file is not in the correct format, a tuple with the value (None, None) can be returned.
        """

        image, path = cls.open(fp), None
        if not image:
            LOGGER.error("Image format not supported.")
            return image, path

        fm, ext = None, None
        if isinstance(image, PngImageFile):
            fm, ext = "PNG", ".png"
            image = image.convert("P", palette=Image.ADAPTIVE)
        elif isinstance(image, JpegImageFile):
            fm, ext = "JPEG", ".jpg"
        elif isinstance(image, WebPImageFile):
            fm, ext = "WEBP", ".webp"

        if app_settings.image_optimizer.to_webp:
            fm, ext = "WEBP", ".webp"

        if str(ext) in cls._supported_file_types:
            image = cls.crop(image, box=box)
            image = cls.resize(image, max_width, max_height)
            image.info = {}
            if not filename and not isinstance(filename, str):
                filename = str(
                    cls.get_identifier(fp.filename if isinstance(fp, _Image) else fp)
                )

            filename += ext
            save_path, path = get_paths(filename, upload_to=upload_to)
            image.save(
                save_path,
                fm,
                optimize=True,
                quality=quality,
                compress_level=compress_level,
            )

            if remove_origin:
                LOGGER.info("Proceed to delete the original image file.")
                if isinstance(fp, (FieldFile, ImageFieldFile)):
                    if filename not in fp.name:
                        fp.delete(save=False)
                else:
                    origin_fp = (
                        getattr(fp, "name", None)
                        if isinstance(fp, (BytesIO, BufferedReader))
                        else fp
                    )
                    if (
                        origin_fp
                        and isinstance(origin_fp, str)
                        and path not in origin_fp
                    ):
                        safe_remove_file(origin_fp)
        else:
            LOGGER.error("Image format not supported.")
        return image, path

    @classmethod
    def crop(cls, image: _Image, box: tuple[int, int, int, int] = None) -> Image:
        """Crop an image

        Args:
          image: PIL image object.
          box: The crop rectangle, as a (left, upper, right, lower)-tuple.

        Returns:
          Returns a rectangular region from this PIL image.
        """
        LOGGER.info("Proceed to crop image.")

        if image and box is not None:
            return image.crop(box)
        return image

    @classmethod
    def resize(
        cls,
        image: _Image,
        width: int = app_settings.image_optimizer.max_width,
        height: int = app_settings.image_optimizer.max_height,
    ) -> Image:
        """Resize image to fit with Width and Height

        Args:
          image: PIL image object.
          width: Max width to resize.
          height: Max height to resize.

        Returns:
          PIL image after resizing
        """
        LOGGER.info("Proceed to reduce image size")

        w, h = image.size
        aspect_ratio = w / h

        if w > width or h > height:
            if aspect_ratio > 1:
                nw = width
                nh = int(width / aspect_ratio)
            else:
                nh = height
                nw = int(height * aspect_ratio)

            return image.resize((nw, nh), Image.LANCZOS)
        return image


MapOptimizer = {TypeChoices.IMAGE: ImageOptimizer}



"""
Django settings for examples project.

Generated by 'django-admin startproject' using Django 5.0.7.

For more information on this file, see
https://docs.djangoproject.com/en/5.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/5.0/ref/settings/
"""

from __future__ import annotations

import os
from pathlib import Path
from random import choice
from string import ascii_letters


# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/5.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-!3lv_tfp9sti+4m611h(=n(ddw54z&8hm(8px(vg+jv5kp^rgq"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ["*"]


# Application definition

DJANGO_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

LOCAL_APPS = [
    "django_chunk_file_upload",
]

INSTALLED_APPS = DJANGO_APPS + LOCAL_APPS

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "examples.wsgi.application"


# Database
# https://docs.djangoproject.com/en/5.0/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}


# Password validation
# https://docs.djangoproject.com/en/5.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/5.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.0/howto/static-files/

STATIC_URL = "static/"

# MEDIA
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#media-root
MEDIA_ROOT = BASE_DIR / "tests/media"
MEDIA_URL = "/media/"

# Default primary key field type
# https://docs.djangoproject.com/en/5.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"


# Django Chunk File Upload Config
UPLOAD_TO_TEMP_DIR = os.path.join(
    MEDIA_ROOT, "".join([choice(ascii_letters) for _ in range(15)])
)
DJANGO_CHUNK_FILE_UPLOAD = {
    "chunk_size": 1024 * 1024 * 2,  # # custom chunk size upload (default: 2MB).
    "upload_to": UPLOAD_TO_TEMP_DIR,
    "is_metadata_storage": True,  # save file metadata,
    "remove_file_on_update": True,
    "optimize": True,
    "image_optimizer": {
        "quality": 82,
        "compress_level": 9,
        "max_width": 1280,
        "max_height": 720,
        "to_webp": True,  # focus convert image to webp type.
    },
    "permission_classes": ("django_chunk_file_upload.permissions.AllowAny",),
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
django-chunk-file-upload
------------

Tests for `django-chunk-file-upload` models module.
"""

import os

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase
from django.test.client import MULTIPART_CONTENT
from django.urls import reverse_lazy

from django_chunk_file_upload import permissions
from django_chunk_file_upload.app_settings import app_settings
from django_chunk_file_upload.models import FileManager
from django_chunk_file_upload.optimize import ImageOptimizer
from django_chunk_file_upload.utils import create_dir, remove_dir
from django_chunk_file_upload.views import ChunkedUploadView


class BaseTestCase(TestCase):
    """Base DjangoChunkFileUpload Test Case"""

    CHUNK_SIZE = 65536
    IMAGE_FILE = os.path.join(settings.BASE_DIR, "tests/media", "test.jpg")
    User = get_user_model()
    origin_image = None
    origin_image_checksum = None
    image = None
    image_path = None
    image_checksum = None
    username = password = "test"

    def setUp(self):
        create_dir(app_settings.upload_to)
        self.origin_image = ImageOptimizer.open(self.IMAGE_FILE)
        self.image, self.image_path = ImageOptimizer.optimize(
            self.IMAGE_FILE,
            filename=None,
            upload_to=app_settings.upload_to,
            max_width=600,
            max_height=600,
            remove_origin=False,
        )
        self.origin_image_checksum = ImageOptimizer.checksum(self.IMAGE_FILE)
        self.image_checksum = ImageOptimizer.checksum(
            os.path.join(settings.MEDIA_ROOT, self.image_path)
        )

    def test_image_optimize(self):
        self.assertNotEqual(
            self.origin_image.size,
            self.image.size,
            "Original image same size with optimized image.",
        )
        self.assertNotEqual(
            self.origin_image_checksum,
            self.image_checksum,
            "Original image checksum same with optimized image checksum.",
        )

    def tearDown(self):
        ImageOptimizer.close(self.origin_image)
        ImageOptimizer.close(self.image)
        remove_dir(app_settings.upload_to)


class TestDjangoChunkUploadView(BaseTestCase):
    file_stat = None
    chunk_from, chunk_to = None, None

    def setUp(self) -> None:
        super().setUp()
        self.file_stat = os.stat(self.IMAGE_FILE)
        self.chunk_from, self.chunk_to = 0, self.CHUNK_SIZE

    def _get_headers(self):
        return {
            "X-File-ID": None,
            "X-File-Name": "test.jpg",
            "X-File-Checksum": self.origin_image_checksum,
            "X-File-Chunk-From": self.chunk_from,
            "X-File-Chunk-Size": self.chunk_to,
            "X-File-Chunk-To": self.chunk_to,
            "X-File-EOF": bool(self.chunk_to > self.file_stat.st_size),
            "X-File-Size": self.file_stat.st_size,
            "X-File-MimeType": "image/jpeg",
        }

    def _get_response(self):
        with open(self.IMAGE_FILE, "rb") as f:
            chunk_from, chunk_to = 0, self.CHUNK_SIZE
            response = None
            while chunk := f.read(self.CHUNK_SIZE):
                response = self.client.post(
                    path=reverse_lazy("django_chunk_file_upload:uploads"),
                    data={"file": SimpleUploadedFile("test.jpg", chunk)},
                    content_type=MULTIPART_CONTENT,
                    headers=self._get_headers(),
                )
                chunk_from += self.CHUNK_SIZE
                chunk_to += self.CHUNK_SIZE

            return response

    def _create_user(self, is_superuser: bool = False, is_staff: bool = False):
        user, _ = self.User.objects.update_or_create(
            username=self.username,
            defaults=dict(
                email="test@example.com",
                password=self.password,
                is_active=True,
                is_staff=is_staff,
                is_superuser=is_superuser,
            ),
        )
        return user

    def _instance_validate(self):
        instance = FileManager.objects.filter(
            checksum=self.origin_image_checksum
        ).first()
        self.assertNotEqual(instance, None, "Not found file manager instance.")
        self.assertEqual(
            instance.checksum, self.origin_image_checksum, "Invalid MD5 checksum."
        )
        instance.delete()

    def test_upload_view_with_allow_any_permission(self):
        """Test ChunkedUploadView with AllowAny permission."""

        ChunkedUploadView.permission_classes = (permissions.AllowAny,)
        response = self._get_response()
        self.assertEqual(201, response.status_code, "Failed to upload files.")
        self._instance_validate()

    def test_upload_view_with_is_authenticated_permission(self):
        """Test ChunkedUploadView with IsAuthenticated permission."""

        ChunkedUploadView.permission_classes = (permissions.IsAuthenticated,)
        response = self._get_response()
        self.assertEqual(
            400, response.status_code, "Failed to test IsAuthenticated permission."
        )
        user = self._create_user()
        self.client.force_login(user)
        response = self._get_response()
        self.assertEqual(
            201, response.status_code, "Failed to test IsAuthenticated permission."
        )
        self._instance_validate()

    def test_upload_view_with_is_superuser_permission(self):
        """Test ChunkedUploadView with IsSuperUser permission."""

        ChunkedUploadView.permission_classes = (permissions.IsSuperUser,)
        response = self._get_response()
        self.assertEqual(
            400, response.status_code, "Failed to test IsSuperUser permission."
        )
        user = self._create_user(is_superuser=True)
        self.client.force_login(user)
        response = self._get_response()
        self.assertEqual(
            201, response.status_code, "Failed to test IsAuthenticated permission."
        )
        self._instance_validate()

    def test_upload_view_with_is_admin_user_permission(self):
        """Test ChunkedUploadView with IsAdminUser permission."""

        ChunkedUploadView.permission_classes = (permissions.IsAdminUser,)
        response = self._get_response()
        self.assertEqual(
            400, response.status_code, "Failed to test IsAdminUser permission."
        )
        user = self._create_user(is_staff=True)
        self.client.force_login(user)
        response = self._get_response()
        self.assertEqual(
            201, response.status_code, "Failed to test IsAuthenticated permission."
        )
        self._instance_validate()

    def test_upload_view_with_is_staff_permission(self):
        """Test uploading file with staff user."""

        ChunkedUploadView.permission_classes = (permissions.IsStaffUser,)
        response = self._get_response()
        self.assertEqual(
            400, response.status_code, "Failed to test IsStaffUser permission."
        )
        user = self._create_user(is_staff=True)
        self.client.force_login(user)
        response = self._get_response()
        self.assertEqual(
            201, response.status_code, "Failed to test IsAuthenticated permission."
        )
        self._instance_validate()


class TestImageOptimizer(BaseTestCase):
    def test_image_optimize(self):
        assert self.origin_image.size != self.image.size
        assert ImageOptimizer.checksum(self.IMAGE_FILE) != ImageOptimizer.checksum(
            os.path.join(settings.MEDIA_ROOT, self.image_path)
        )
